from __future__ import annotations

from collections.abc import Coroutine, Generator
from typing import Any, Callable, Literal, TypeVar, Union

from typing_extensions import Concatenate, ParamSpec, TypeAlias

from resonate.actions import (
    DI,
    LFC,
    LFI,
    RFC,
    RFI,
)
from resonate.context import Context
from resonate.promise import Promise

T = TypeVar("T")
P = ParamSpec("P")


Yieldable: TypeAlias = Union[
    LFI,
    LFC,
    RFI,
    RFC,
    DI,
    Promise[Any],
]

DurableCoro: TypeAlias = Callable[Concatenate[Context, P], Generator[Yieldable, Any, T]]
DurableSyncFn: TypeAlias = Callable[Concatenate[Context, P], T]
DurableAsyncFn: TypeAlias = Callable[Concatenate[Context, P], Coroutine[Any, Any, T]]
DurableFn: TypeAlias = Union[DurableSyncFn[P, T], DurableAsyncFn[P, T]]

Headers: TypeAlias = Union[dict[str, str], None]
Tags: TypeAlias = Union[dict[str, str], None]
IdempotencyKey: TypeAlias = Union[str, None]
Data: TypeAlias = Union[str, None]
State: TypeAlias = Literal[
    "PENDING", "RESOLVED", "REJECTED", "REJECTED_CANCELED", "REJECTED_TIMEDOUT"
]



# This is a copy from library https://github.com/rustedpy/result.
from __future__ import annotations

import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    NoReturn,
    TypeVar,
    Union,
)

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


if TYPE_CHECKING:
    from collections.abc import Iterator


T_co = TypeVar("T_co", covariant=True)  # Success type
E_co = TypeVar("E_co", covariant=True)  # Error type


class Ok(Generic[T_co]):
    """
    A value that indicates success and which stores arbitrary data for the return value.
    """

    __match_args__ = ("ok_value",)
    __slots__ = ("_value",)

    def __iter__(self) -> Iterator[T_co]:
        yield self._value

    def __init__(self, value: T_co) -> None:
        self._value = value

    def __repr__(self) -> str:
        return f"Ok({self._value!r})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Ok) and self._value == other._value

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __hash__(self) -> int:
        return hash((True, self._value))

    def ok(self) -> T_co:
        """
        Return the value.
        """
        return self._value

    def err(self) -> None:
        """
        Return `None`.
        """
        return

    def unwrap(self) -> T_co:
        """
        Return the value.
        """
        return self._value


class DoExceptionError(Exception):
    """
    This is used to signal to `do()` that the result is an `Err`,
    which short-circuits the generator and returns that Err.
    Using this exception for control flow in `do()` allows us
    to simulate `and_then()` in the Err case: namely, we don't call `op`,
    we just return `self` (the Err).
    """

    def __init__(self, err: Err[E_co]) -> None:
        self.err = err


class Err(Generic[E_co]):
    """
    A value that signifies failure and which stores arbitrary data for the error.
    """

    __match_args__ = ("err_value",)
    __slots__ = ("_value",)

    def __iter__(self) -> Iterator[NoReturn]:
        def _iter() -> Iterator[NoReturn]:
            raise DoExceptionError(self)

        return _iter()

    def __init__(self, value: E_co) -> None:
        self._value = value

    def __repr__(self) -> str:
        return f"Err({self._value!r})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Err) and self._value == other._value

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __hash__(self) -> int:
        return hash((False, self._value))

    def ok(self) -> None:
        """
        Return `None`.
        """
        return

    def err(self) -> E_co:
        """
        Return the error.
        """
        return self._value

    def unwrap(self) -> NoReturn:
        """
        Raises an `UnwrapError`.
        """
        exc = UnwrapError(
            self,
            f"Called `Result.unwrap()` on an `Err` value: {self._value!r}",
        )
        if isinstance(self._value, BaseException):
            raise exc from self._value
        raise exc


# define Result as a generic type alias for use
# in type annotations
"""
A simple `Result` type inspired by Rust.
Not all methods (https://doc.rust-lang.org/std/result/enum.Result.html)
have been implemented, only the ones that make sense in the Python context.
"""
Result: TypeAlias = Union[Ok[T_co], Err[E_co]]

"""
A type to use in `isinstance` checks.
This is purely for convenience sake, as you could also just write `isinstance(res, (Ok, Err))`
"""  # noqa: E501


class UnwrapError(Exception):
    """
    Exception raised from ``.unwrap_<...>`` and ``.expect_<...>`` calls.

    The original ``Result`` can be accessed via the ``.result`` attribute, but
    this is not intended for regular use, as type information is lost:
    ``UnwrapError`` doesn't know about both ``T_co`` and ``E_co``, since it's raised
    from ``Ok()`` or ``Err()`` which only knows about either ``T_co`` or ``E_co``,
    not both.
    """

    _result: Result[object, object]

    def __init__(self, result: Result[object, object], message: str) -> None:
        self._result = result
        super().__init__(message)

    @property
    def result(self) -> Result[Any, Any]:
        """
        Returns the original result.
        """
        return self._result



from __future__ import annotations

import os
import time
from threading import Thread
from typing import Any

import requests

from resonate import utils
from resonate.cmd_queue import Claim, CommandQ
from resonate.encoders import JsonEncoder
from resonate.logging import logger
from resonate.stores.record import TaskRecord
from resonate.task_sources.traits import ITaskSource


class Poller(ITaskSource):
    def __init__(
        self,
        url: str | None = None,
        group: str | None = None,
    ) -> None:
        self._url = (
            url
            if url is not None
            else os.getenv("RESONATE_POLLER", "http://localhost:8002")
        )
        self._group = (
            group if group is not None else os.getenv("RESONATE_GROUP", "default")
        )
        self._encoder = JsonEncoder()
        self._t: Thread | None = None

    def start(self, cmd_queue: CommandQ, pid: str) -> None:
        assert self._t is None
        self._t = Thread(target=self._run, args=(cmd_queue, pid), daemon=True)
        self._t.start()

    def stop(self) -> None:
        raise NotImplementedError

    @utils.exit_on_exception
    def _run(self, cmd_queue: CommandQ, pid: str) -> None:
        url = f"{self._url}/{self._group}/{pid}"

        while True:
            try:
                with requests.get(url, stream=True) as res:  # noqa: S113
                    if not res.ok:
                        break

                    for line in res.iter_lines(chunk_size=None, decode_unicode=True):
                        if not line:
                            continue

                        stripped = line.strip()
                        if not stripped.startswith("data:"):
                            continue

                        info = self._encoder.decode(stripped[5:])
                        if "task" not in info:
                            continue

                        # extract the task
                        task = TaskRecord.decode(info["task"], encoder=self._encoder)

                        # enqueue the task
                        cmd_queue.put(Claim(task))

            except requests.exceptions.ConnectionError:
                logger.warning("Connection to poller failed, reconnecting")

            time.sleep(1)

    def default_recv(self, pid: str) -> dict[str, Any]:
        return {"type": "poll", "data": {"group": self._group, "id": pid}}



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypeVar, final

from typing_extensions import Self

from resonate.dataclasses import DurablePromise
from resonate.options import Options

if TYPE_CHECKING:
    from resonate import retry_policy
    from resonate.dataclasses import Invocation


T = TypeVar("T")
_ASSERT_MSG = (
    "Creating durable promises directly does not support adding options this way"
)


@final
@dataclass
class RFI:
    unit: Invocation[Any] | DurablePromise
    opts: Options = field(default=Options())

    def options(
        self,
        *,
        id: str | None = None,
        send_to: str | None = None,
        execute_here: bool = False,
    ) -> Self:
        assert not isinstance(self.unit, DurablePromise), _ASSERT_MSG
        self.opts = Options(id=id, send_to=send_to, execute_here=execute_here)
        return self


@final
@dataclass
class RFC:
    unit: Invocation[Any] | DurablePromise
    opts: Options = field(default=Options())

    def options(
        self,
        *,
        id: str | None = None,
        send_to: str | None = None,
        execute_here: bool = False,
    ) -> Self:
        assert not isinstance(self.unit, DurablePromise), _ASSERT_MSG
        self.opts = Options(id=id, send_to=send_to, execute_here=execute_here)
        return self

    def to_rfi(self) -> RFI:
        return RFI(self.unit, self.opts)


@final
@dataclass
class LFC:
    unit: Invocation[Any]
    opts: Options = field(default=Options())

    def options(
        self,
        *,
        id: str | None = None,
        durable: bool = True,
        retry_policy: retry_policy.RetryPolicy | None = None,
    ) -> Self:
        self.opts = Options(durable=durable, id=id, retry_policy=retry_policy)
        return self

    def to_lfi(self) -> LFI:
        return LFI(self.unit, opts=self.opts)


@final
@dataclass
class DI:
    """
    Dataclass that contains all required information to do a
    deferred invocation.
    """

    id: str
    unit: Invocation[Any]
    opts: Options = field(default=Options())

    def options(self) -> Self:
        self.opts = Options(durable=True, id=self.id)
        return self


@final
@dataclass
class LFI:
    unit: Invocation[Any]
    opts: Options = field(default=Options())

    def options(
        self,
        *,
        id: str | None = None,
        durable: bool = True,
        retry_policy: retry_policy.RetryPolicy | None = None,
    ) -> Self:
        self.opts = Options(durable=durable, id=id, retry_policy=retry_policy)
        return self



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, final

from resonate.errors import ResonateError
from resonate.stores.record import (
    DurablePromiseRecord,
    Param,
    TaskRecord,
    Value,
)
from resonate.stores.traits import IPromiseStore
from resonate.time import now

if TYPE_CHECKING:
    from collections.abc import Hashable

    from resonate.typing import Data, Headers, State, Tags


def _timeout(promise_record: DurablePromiseRecord) -> DurablePromiseRecord:
    new_status: State = "REJECTED_TIMEDOUT"
    if promise_record.is_pending() and now() >= promise_record.timeout:
        if promise_record.tags is not None:
            resonate_timeout: Hashable | None = promise_record.tags.get(
                "resonate:timeout"
            )
            if resonate_timeout is not None and resonate_timeout == "true":
                new_status = "RESOLVED"
        return DurablePromiseRecord(
            new_status,
            id=promise_record.id,
            timeout=promise_record.timeout,
            param=promise_record.param,
            value=Value(headers=None, data=None),
            created_on=promise_record.created_on,
            completed_on=promise_record.timeout,
            idempotency_key_for_create=promise_record.idempotency_key_for_create,
            idempotency_key_for_complete=None,
            tags=promise_record.tags,
        )
    return promise_record


class IStorage(ABC):
    @abstractmethod
    def rmw(
        self,
        id: str,
        fn: Callable[[DurablePromiseRecord | None], DurablePromiseRecord],
    ) -> DurablePromiseRecord: ...


class MemoryStorage(IStorage):
    def __init__(self) -> None:
        self._durable_promises: dict[str, DurablePromiseRecord] = {}

    def rmw(
        self,
        id: str,
        fn: Callable[[DurablePromiseRecord | None], DurablePromiseRecord],
    ) -> DurablePromiseRecord:
        promise_record = self._durable_promises.get(id)
        if promise_record is not None:
            promise_record = _timeout(promise_record)
        item = fn(promise_record)
        self._durable_promises[id] = item
        return item


@final
class LocalPromiseStore(IPromiseStore):
    def __init__(self, storage: IStorage | None = None) -> None:
        self._storage = storage or MemoryStorage()

    def create_with_task(  # noqa: PLR0913
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
        pid: str,
        ttl: int,
    ) -> tuple[DurablePromiseRecord, TaskRecord | None]:
        return self.create(
            id=id,
            ikey=ikey,
            strict=strict,
            headers=headers,
            data=data,
            timeout=timeout,
            tags=tags,
        ), None

    def create(  # noqa: PLR0913
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
    ) -> DurablePromiseRecord:
        def _create(
            promise_record: DurablePromiseRecord | None,
        ) -> DurablePromiseRecord:
            if promise_record is None:
                return DurablePromiseRecord(
                    state="PENDING",
                    id=id,
                    timeout=timeout,
                    param=Param(headers=headers, data=data),
                    value=Value(headers=None, data=None),
                    created_on=now(),
                    completed_on=None,
                    idempotency_key_for_complete=None,
                    idempotency_key_for_create=ikey,
                    tags=tags,
                )
            if strict and not promise_record.is_pending():
                msg = "Forbidden request: Durable promise previously created"
                raise ResonateError(
                    msg,
                    "STORE_FORBIDDEN",
                )
            if (
                promise_record.idempotency_key_for_create is None
                or ikey != promise_record.idempotency_key_for_create
            ):
                msg = "Forbidden request: Missing idempotency key for create"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            return promise_record

        return self._storage.rmw(id=id, fn=_create)

    def reject(
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        def _reject(
            promise_record: DurablePromiseRecord | None,
        ) -> DurablePromiseRecord:
            if promise_record is None:
                msg = "Not found"
                raise ResonateError(msg, "STORE_NOT_FOUND")
            if promise_record.is_pending():
                return DurablePromiseRecord(
                    state="REJECTED",
                    id=promise_record.id,
                    timeout=promise_record.timeout,
                    param=promise_record.param,
                    value=Value(headers=headers, data=data),
                    created_on=promise_record.created_on,
                    completed_on=now(),
                    idempotency_key_for_create=promise_record.idempotency_key_for_create,
                    idempotency_key_for_complete=ikey,
                    tags=promise_record.tags,
                )
            if strict and not promise_record.is_rejected():
                msg = "Forbidden request"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            if not promise_record.is_timeout() and (
                promise_record.idempotency_key_for_complete is None
                or ikey != promise_record.idempotency_key_for_complete
            ):
                msg = "Forbidden request"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            return promise_record

        return self._storage.rmw(id=id, fn=_reject)

    def cancel(
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        def _cancel(
            promise_record: DurablePromiseRecord | None,
        ) -> DurablePromiseRecord:
            if promise_record is None:
                msg = "Not Found"
                raise ResonateError(msg, "STORE_NOT_FOUND")
            if promise_record.is_pending():
                return DurablePromiseRecord(
                    state="REJECTED_CANCELED",
                    id=promise_record.id,
                    timeout=promise_record.timeout,
                    param=promise_record.param,
                    value=Value(headers=headers, data=data),
                    created_on=promise_record.created_on,
                    completed_on=now(),
                    idempotency_key_for_create=promise_record.idempotency_key_for_create,
                    idempotency_key_for_complete=ikey,
                    tags=promise_record.tags,
                )
            if strict and not promise_record.is_canceled():
                msg = "Forbidden request"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            if not promise_record.is_timeout() and (
                promise_record.idempotency_key_for_complete is None
                or ikey != promise_record.idempotency_key_for_complete
            ):
                msg = "Forbidden request"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            return promise_record

        return self._storage.rmw(id=id, fn=_cancel)

    def resolve(
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        def _resolve(
            promise_record: DurablePromiseRecord | None,
        ) -> DurablePromiseRecord:
            if promise_record is None:
                msg = "Not found"
                raise ResonateError(msg, "STORE_NOT_FOUND")
            if promise_record.is_pending():
                return DurablePromiseRecord(
                    state="RESOLVED",
                    id=promise_record.id,
                    timeout=promise_record.timeout,
                    param=promise_record.param,
                    value=Value(headers=headers, data=data),
                    created_on=promise_record.created_on,
                    completed_on=now(),
                    idempotency_key_for_create=promise_record.idempotency_key_for_create,
                    idempotency_key_for_complete=ikey,
                    tags=promise_record.tags,
                )
            if strict and not promise_record.is_resolved():
                msg = "Forbidden request"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            if not promise_record.is_timeout() and (
                promise_record.idempotency_key_for_complete is None
                or ikey != promise_record.idempotency_key_for_complete
            ):
                msg = "Forbidden request"
                raise ResonateError(msg, "STORE_FORBIDDEN")
            return promise_record

        return self._storage.rmw(id=id, fn=_resolve)

    def get(self, *, id: str) -> DurablePromiseRecord:
        def _get(
            promise_record: DurablePromiseRecord | None,
        ) -> DurablePromiseRecord:
            if promise_record is None:
                raise ResonateError(msg="Not Found", code="STORE_NOT_FOUND")

            return promise_record

        return self._storage.rmw(id=id, fn=_get)


@final
class LocalStore:
    def __init__(self, promises: LocalPromiseStore | None = None) -> None:
        self.promises = promises or LocalPromiseStore()



from __future__ import annotations

from dataclasses import dataclass
from queue import Queue
from typing import TYPE_CHECKING, Any, Union

from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from resonate.dataclasses import Invocation
    from resonate.handle import Handle
    from resonate.result import Result
    from resonate.stores.record import TaskRecord


@dataclass(frozen=True)
class Invoke:
    id: str


@dataclass(frozen=True)
class ForkOrJoin:
    id: str
    handle: Handle[Any]
    invocation: Invocation[Any]


@dataclass(frozen=True)
class Subscribe:
    id: str
    handle: Handle[Any]


@dataclass(frozen=True)
class Notify:
    id: str
    value: Result[Any, Exception]


@dataclass(frozen=True)
class Resume:
    id: str
    next_value: Result[Any, Exception]


@dataclass(frozen=True)
class Complete:
    id: str
    result: Result[Any, Exception]


@dataclass(frozen=True)
class Claim:
    record: TaskRecord


Command: TypeAlias = Union[
    Invoke, Resume, Complete, Claim, Subscribe, Notify, ForkOrJoin
]
CommandQ: TypeAlias = Queue[Union[Command, None]]



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar, final

T = TypeVar("T")


@final
@dataclass(frozen=True)
class Promise(Generic[T]):
    id: str



from __future__ import annotations

from threading import Thread
from typing import TYPE_CHECKING, Any

from resonate import utils
from resonate.cmd_queue import CommandQ, Complete
from resonate.delay_queue import Queue
from resonate.processor.traits import IProcessor
from resonate.result import Err, Ok, Result

if TYPE_CHECKING:
    from resonate.processor.traits import SQE


class Processor(IProcessor):
    def __init__(
        self,
        workers: int = 4,
    ) -> None:
        assert workers > 0, "workers must be greater than zero."

        self._workers = workers
        self._threads = set[Thread]()

        self._sq: Queue[SQE[Any]] = Queue()

    def start(self, cmd_queue: CommandQ, pid: str) -> None:
        for _ in range(self._workers):
            t = Thread(target=self._run, args=(cmd_queue,), daemon=True)
            self._threads.add(t)

            t.start()

    def stop(self) -> None:
        raise NotImplementedError

    def enqueue(self, sqe: SQE[Any]) -> None:
        self._sq.put(sqe)

    @utils.exit_on_exception
    def _run(self, cmd_queue: CommandQ) -> None:
        while True:
            sqe = self._sq.get()
            result: Result[Any, Exception]

            try:
                result = Ok(sqe.thunk())
            except Exception as e:  # noqa: BLE001
                result = Err(e)

            # put on completion queue
            cmd_queue.put(Complete(sqe.id, result))



from __future__ import annotations

from typing import TYPE_CHECKING, final

from opentelemetry import context, trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from typing_extensions import assert_never

from resonate.events import (
    ExecutionAwaited,
    ExecutionInvoked,
    ExecutionResumed,
    ExecutionTerminated,
    PromiseCompleted,
    PromiseCreated,
    SchedulerEvents,
)
from resonate.tracing import IAdapter

if TYPE_CHECKING:
    from opentelemetry.trace import Tracer
    from opentelemetry.trace.span import Span


@final
class OpenTelemetryAdapter(IAdapter):
    def __init__(self, app_name: str, endpoint: str) -> None:
        self._spans: dict[str, tuple[Span, object]] = {}
        assert __package__ is not None

        provider = TracerProvider(resource=Resource.create({SERVICE_NAME: app_name}))
        exporter = OTLPSpanExporter(endpoint=endpoint)
        processor = BatchSpanProcessor(exporter)
        provider.add_span_processor(processor)
        trace.set_tracer_provider(provider)

        self._tracer: Tracer = trace.get_tracer(
            __package__,
        )

    def _get_span(self, id: str) -> Span:
        return self._spans[id][0]

    def process_event(self, event: SchedulerEvents) -> None:
        if isinstance(event, PromiseCreated):
            assert (
                event.id not in self._spans
            ), "There shouldn't be never another span with same name."
            self._create_span(event.id, event.parent_id, start_time=event.tick)
        elif isinstance(event, PromiseCompleted):
            assert (
                event.id in self._spans
            ), "There should always be an span with that name."
            self._close_span(event.id, end_time=event.tick)
        elif isinstance(event, ExecutionInvoked):
            span = self._get_span(event.id)
            span.add_event(
                ExecutionInvoked.__name__,
            )
        elif isinstance(event, ExecutionTerminated):
            span = self._get_span(event.id)
            span.add_event(
                ExecutionTerminated.__name__,
                timestamp=event.tick,
            )
        elif isinstance(event, ExecutionResumed):
            span = self._get_span(event.id)

            span.add_event(
                ExecutionResumed.__name__,
                timestamp=event.tick,
            )
        elif isinstance(event, ExecutionAwaited):
            span = self._get_span(event.id)
            span.add_event(
                ExecutionAwaited.__name__,
                timestamp=event.tick,
            )
        else:
            assert_never(event)

    def _create_span(self, id: str, parent_id: str | None, start_time: int) -> None:
        parent_span_and_token = (
            self._spans[parent_id] if parent_id is not None else None
        )
        parent_span: Span | None = None
        if parent_span_and_token is not None:
            parent_span = parent_span_and_token[0]

        parent_ctx = (
            trace.set_span_in_context(parent_span) if parent_span is not None else None
        )
        token = context.attach(parent_ctx) if parent_ctx is not None else None
        new_span = self._tracer.start_span(
            name=id, context=parent_ctx, start_time=start_time
        )

        assert (
            id not in self._spans
        ), "There should not be two spans with the same name at the same time."
        self._spans[id] = (new_span, token)

    def _close_span(self, id: str, end_time: int) -> None:
        assert (
            id in self._spans
        ), "There should be an span associated with the promise id."
        span, token = self._spans.pop(id)
        if token:
            context.detach(token=token)
        span.end(end_time=end_time)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from resonate.cmd_queue import CommandQ


class Subsystem(ABC):
    @abstractmethod
    def start(self, cmd_queue: CommandQ, pid: str) -> None: ...

    @abstractmethod
    def stop(self) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, TypeVar, final, overload

from typing_extensions import Concatenate, ParamSpec

from resonate.actions import (
    DI,
    LFC,
    LFI,
    RFC,
    RFI,
)
from resonate.dataclasses import DurablePromise, Invocation, RegisteredFn
from resonate.time import now

if TYPE_CHECKING:
    from collections.abc import Coroutine, Generator

    from resonate.dependencies import Dependencies
    from resonate.typing import DurableCoro, DurableFn, Yieldable

P = ParamSpec("P")
T = TypeVar("T")


@final
class Context:
    def __init__(
        self,
        deps: Dependencies,
    ) -> None:
        self._deps = deps

    def get_dependency(self, key: str) -> Any:  # noqa: ANN401
        return self._deps.get(key)

    def sleep(self, secs: int) -> RFC:
        return self.rfc(
            DurablePromise(
                tags={"resonate:timeout": "true"}, timeout=now() + (secs * 1_000)
            )
        )

    @overload
    def rfc(self, cmd: DurablePromise, /) -> RFC: ...
    @overload
    def rfc(
        self,
        func: RegisteredFn[P, Any],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFC: ...
    @overload
    def rfc(self, func: str, /, *args: Any, **kwargs: Any) -> RFC: ...  # noqa: ANN401
    @overload
    def rfc(
        self,
        func: Callable[Concatenate[Context, P], Generator[Yieldable, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFC: ...
    @overload
    def rfc(
        self,
        func: Callable[Concatenate[Context, P], Any | Coroutine[Any, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFC: ...
    def rfc(
        self,
        func_or_cmd: DurableCoro[P, T]
        | DurableFn[P, T]
        | str
        | DurablePromise
        | RegisteredFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFC:
        unit: Invocation[Any] | DurablePromise
        if isinstance(func_or_cmd, str):
            unit = Invocation(func_or_cmd, *args, **kwargs)
        elif isinstance(func_or_cmd, DurablePromise):
            unit = func_or_cmd
        elif isinstance(func_or_cmd, RegisteredFn):
            unit = Invocation(func_or_cmd.fn, *args, **kwargs)
        else:
            unit = Invocation(func_or_cmd, *args, **kwargs)
        return RFC(unit)

    @overload
    def rfi(self, cmd: DurablePromise, /) -> RFI: ...
    @overload
    def rfi(
        self,
        func: RegisteredFn[P, Any],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFI: ...
    @overload
    def rfi(self, func: str, /, *args: Any, **kwargs: Any) -> RFI: ...  # noqa: ANN401
    @overload
    def rfi(
        self,
        func: Callable[Concatenate[Context, P], Generator[Yieldable, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFI: ...
    @overload
    def rfi(
        self,
        func: Callable[Concatenate[Context, P], Any | Coroutine[Any, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFI: ...
    def rfi(
        self,
        func_or_cmd: DurableCoro[P, T]
        | DurableFn[P, T]
        | str
        | DurablePromise
        | RegisteredFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RFI:
        return self.rfc(func_or_cmd, *args, **kwargs).to_rfi()

    @overload
    def lfi(
        self,
        func: RegisteredFn[P, Any],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFI: ...
    @overload
    def lfi(
        self,
        func: Callable[Concatenate[Context, P], Generator[Yieldable, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFI: ...
    @overload
    def lfi(
        self,
        func: Callable[Concatenate[Context, P], Any | Coroutine[Any, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFI: ...
    def lfi(
        self,
        func_or_cmd: DurableCoro[P, T] | DurableFn[P, T] | RegisteredFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFI:
        """
        Local function invocation.

        Invoke and immediatelly receive a `Promise[T]` that
        represents the future result of the execution.

        The `Promise` can be yielded later in the execution to await
        for the result.
        """
        return self.lfc(func_or_cmd, *args, **kwargs).to_lfi()

    @overload
    def lfc(
        self,
        func: RegisteredFn[P, Any],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFC: ...
    @overload
    def lfc(
        self,
        func: Callable[Concatenate[Context, P], Generator[Yieldable, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFC: ...
    @overload
    def lfc(
        self,
        func: Callable[Concatenate[Context, P], Any | Coroutine[Any, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFC: ...
    def lfc(
        self,
        func_or_cmd: DurableCoro[P, T] | DurableFn[P, T] | RegisteredFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> LFC:
        """
        Local function call.

        LFC and await for the result of the execution. It's syntax
        sugar for `yield (yield ctx.lfi(...))`
        """
        unit: Invocation[Any]
        if isinstance(func_or_cmd, RegisteredFn):
            unit = Invocation(func_or_cmd.fn, *args, **kwargs)
        else:
            unit = Invocation(func_or_cmd, *args, **kwargs)
        return LFC(unit)

    @overload
    def deferred(
        self,
        id: str,
        coro: Callable[Concatenate[Context, P], Generator[Yieldable, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> DI: ...
    @overload
    def deferred(
        self,
        id: str,
        coro: Callable[Concatenate[Context, P], Any | Coroutine[Any, Any, Any]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> DI: ...
    def deferred(
        self,
        id: str,
        coro: DurableCoro[P, T] | DurableFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> DI:
        """
        Deferred invocation.

        Invoke as a root invocation. Is equivalent to do `Scheduler.run(...)`
        invoked execution will be retried and managed from the server.
        """
        return DI(id=id, unit=Invocation(coro, *args, **kwargs))
